/******************************************************************************
 * macro.c
 * SASS Compiler for CoreTone, sfx + instrument parsing and pack generation.
 *-----------------------------------------------------------------------------
 * Copyright (C) 2015 - 2016 Osman Celimli
 * For conditions of distribution and use, see copyright notice in ctcc.c
 ******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdint.h>

#include "../types.h"
#include "../CoreTone/sample.h"
#include "../CoreTone/channel.h"
#include "../CoreTone/music.h"

#include "useful.h"
#include "sample.h"
#include "macro.h"

/******************************************************************************
 * Text Tables
 ******************************************************************************/
extern const char szERR_NoMainMem[];
extern const char szERR_Null[];
extern const char szERR_Range[];
extern const char szERR_Syntax[];
extern const char szERR_EarlyEnd[];
extern const char szERR_NoOpen[];
extern const char szERR_NoDest[];

extern const char szERR_Garbage[];

const char szMAC_Parse[] =		"Parsing macro script file : ";
const char szINS_Make[] =		"Saving instrument package to : ";
const char szSFX_Make[] =		"Saving sound effect binary to : ";

const char szMAC_Empty[] =		"WARNING- No macros defined, are you sure about this?\n";
const char szMAC_PickSample[] =	"WARNING- No sample set for channel of : ";
const char szMAC_SetOff[] =		"WARNING- No note off point defined for instrument : ";
const char szMAC_OffSFX[] =		"ERROR- Note off assigned for sound effect : ";
const char szMAC_MultiOff[] =	"ERROR- Multiple note off assignments for : ";

const char szMAC_NoCom[] =		"ERROR- Invalid command : ";
const char szMAC_NoFlow[] =		"ERROR- Flow command outside of script header : ";
const char szMAC_EarlyMode[] =	"ERROR- Sample mode cannot be set until a sample has been assigned\n";
const char szMAC_NoMode[] =		"ERROR- Invalid sampling mode : ";
const char szMAC_LoopStart[] =	"ERROR- Sample loop missing start offset\n";
const char szMAC_LoopEnd[] =	"ERROR- Sample loop missing end offset\n";
const char szMAC_LoopRange[] =	"ERROR- Sample loop end offset must be greater than the start\n";

const char szMAC_BadLabel[] =	"ERROR- Invalid macro label in : ";
const char szMAC_BadPri[] =		"ERROR- Priority out of range in : ";
const char szMAC_NoPri[] =		"WARNING- No priority specified for sound effect : ";
const char szMAC_InstPri[] =	"ERROR- Priority assigned to instrument in : ";
const char szMAC_TwoLabels[] =	"ERROR- Macro already defined with label : ";
const char szMAC_Unclear[] =	"ERROR- Unclear macro syntax in : ";
const char szMAC_InsMultiC[] =	"ERROR- Instrument uses more than one channel in : ";
const char szMAC_NoSample[] =	"ERROR- Unable to locate sample with name : ";
const char szMAC_ManySample[] =	"ERROR- Multiple sample assignments to one channel of : ";
const char szMAC_Busy[] =		"ERROR- Unexpected end of macro in : ";

const char szMAC_BadWrite[] =	"ERROR- Couldn't write to output file : ";
const char szMAC_BadDef[] =		"ERROR- Can't create priority definition file\n";

const char szMAC_WhiteSpace[] =	" ,\t\n";
const char szMAC_DirMarkers[] =	"\\/>";
const char szMAC_SfxExt[] =		".sfx";
const char szMAC_DefExt[] =		".h";

/******************************************************************************
 * Globals and Externs
 ******************************************************************************/
/* Parser State
 *----------------------------------------------------------------------------*/
char *pszMacroFile = NULL;
char *pszMacroBuf = NULL;
uint32_t uiMacroLine = 0;
int32_t giNoSamples = 0;

/* Macro Sound Effect / Instrument Table
 *----------------------------------------------------------------------------*/
MacroBlock_t *pMacros = NULL;
uint32_t uiMacros = 0, uiMacroSlots = 0;


/******************************************************************************
 * !!!!----             MACRO INSTRUMENT / SFX COMMANDS              ----!!!!
 ******************************************************************************/
typedef int32_t (*MacroCom_t)(MacroBlock_t *pMacro, MacroChannel_t *pChannel);

/* sample,s name
 -----------------------------------------------------------------------------*/
int32_t macCom_sample(MacroBlock_t *pMacro, MacroChannel_t *pChannel)
{
	char *pszSubTok;

	Sample_t *pSample;
	uint32_t uiSampleNum;

	if(NULL == (pszSubTok = strtok(NULL, szMAC_WhiteSpace)))
		return 1;
	if(!giNoSamples)
	{
		if(smp_get(pszSubTok, &uiSampleNum, &pSample))
		{
			fprintf(stderr, "%s%s in : %s,%u\n", szMAC_NoSample,
				pszSubTok, pszMacroFile, uiMacroLine);
			return 1;
		}
	}
	else
	{
		uiSampleNum = -1;
	}

	if(pChannel->iGotSample)
	{
		fprintf(stderr, "%s%s in : %s,%u\n", szMAC_ManySample,
			pMacro->pszName, pszMacroFile, uiMacroLine);
		return 1;
	}

	pChannel->uiSample = uiSampleNum;
	pChannel->iGotSample = -1;
	return 0;
}

/* mode,m singleshot/looping [start] [end]
 -----------------------------------------------------------------------------*/
int32_t macCom_mode(MacroBlock_t *pMacro, MacroChannel_t *pChannel)
{
	char *pszModeTok,*pszStartTok,*pszEndTok;
	int16p16_t stStart,stEnd;
	int8p8_t stWrite;

	int32_t iStatus;


	/**
	 * Assigning a sample to the channel prior to selecting a mode isn't
	 * technically required by the driver since the sample is only
	 * set once in the channel header, but we may perform checks
	 * in regard to the chosen sample a bit further below...
	 */
	if(!(pChannel->iGotSample))
	{
		fprintf(stderr, "%s", szMAC_EarlyMode);
		return 1;
	}

	pszModeTok = strtok(NULL, szMAC_WhiteSpace);
	if(0 == strcmp(pszModeTok, "singleshot"))
	{
		iStatus = mac_addByte(pChannel, CORETONE_PATCH_MODE_SINGLESHOT);
	}
	else if(0 == strcmp(pszModeTok, "looping"))
	{
		pszStartTok = strtok(NULL, szMAC_WhiteSpace);
		pszEndTok = strtok(NULL, szMAC_WhiteSpace);
		if(NULL == pszStartTok)
		{
			fprintf(stderr, "%s", szMAC_LoopStart);
			iStatus = 1;
		}
		else if(NULL == pszEndTok)
		{
			fprintf(stderr, "%s", szMAC_LoopEnd);
			iStatus = 1;
		}
		else
		{
			/**
			 * ---- Some thought in regards to sample range checking :
			 *
			 *  While an invalid sample loop range could cause the waveform
			 * renderer to wander off into some nasty swamp of memory, the
			 * same behavior can be used to bridge adjacent samples.
			 *  A warning let's novices know that they're doing something
			 * which may be considered dangerous, but won't stop an
			 * experienced user from (ab)using interesting behavior. I'll
			 * leave this out for now since the sample lengths aren't
			 * actually yanked from the descriptor.
			 *  However, the offsets DO need to be capped at the maximum
			 * size of a sample and verified that uiStart < uiEnd.
			 *
			 * Anyway - the start offset is INCLUSIVE while the end offset
			 * is EXCLUSIVE (renderer just calculates the segment length).
			 */
			if(parse16p16(&stStart, pszStartTok))
				return 1;
			if(parse16p16(&stEnd, pszEndTok))
				return 1;

			if((stStart.usPair.usHi >= CORETONE_SAMPLES_MAXLENGTH)
				|| (stEnd.usPair.usHi > CORETONE_SAMPLES_MAXLENGTH))
			{
				fprintf(stderr, "%s", szERR_Range);
				return 1;
			}
			else if(stStart.usPair.usHi >= stEnd.usPair.usHi)
			{
				fprintf(stderr, "%s", szMAC_LoopRange);
				return 1;
			}

			iStatus = mac_addByte(pChannel, CORETONE_PATCH_MODE_LOOP);
			stWrite.usWhole = stStart.usPair.usHi;
			iStatus |= mac_addByte(pChannel, stWrite.ucPair.ucLo);
			iStatus |= mac_addByte(pChannel, stWrite.ucPair.ucHi);
			stWrite.usWhole = stEnd.usPair.usHi;
			iStatus |= mac_addByte(pChannel, stWrite.ucPair.ucLo);
			iStatus |= mac_addByte(pChannel, stWrite.ucPair.ucHi);
		}
	}
	else
	{
		if(NULL != pszModeTok)
		{
			fprintf(stderr, "%s%s\n", szMAC_NoMode, pszModeTok);
		}
		iStatus = 1;
	}

	return iStatus;
}

/* frequency,f offset [adjustment]
 -----------------------------------------------------------------------------*/
int32_t macCom_frequency(MacroBlock_t *pMacro, MacroChannel_t *pChannel)
{
	char *pszOffTok,*pszAdjTok;
	int16p16_t stOff,stAdj;
	int8p8_t stWrite;

	int32_t iStatus;


	pszOffTok = strtok(NULL, szMAC_WhiteSpace);
	pszAdjTok = strtok(NULL, szMAC_WhiteSpace);
	if(NULL == pszOffTok)
		return 1;
	if(parse16p16(&stOff, pszOffTok))
		return 1;

	if(NULL != pszAdjTok)
	{
		if(parse16p16(&stAdj, pszAdjTok))
		{
			return 1;
		}
	}
	else
	{
		stAdj.iWhole = 0;
	}

	iStatus = mac_addByte(pChannel, CORETONE_PATCH_FREQUENCY);
	stWrite.usWhole = stOff.usPair.usLo;
	iStatus |= mac_addByte(pChannel, stWrite.ucPair.ucLo);
	iStatus |= mac_addByte(pChannel, stWrite.ucPair.ucHi);
	stWrite.usWhole = stOff.usPair.usHi;
	iStatus |= mac_addByte(pChannel, stWrite.ucPair.ucLo);
	iStatus |= mac_addByte(pChannel, stWrite.ucPair.ucHi);

	stWrite.usWhole = stAdj.usPair.usLo;
	iStatus |= mac_addByte(pChannel, stWrite.ucPair.ucLo);
	iStatus |= mac_addByte(pChannel, stWrite.ucPair.ucHi);
	stWrite.usWhole = stAdj.usPair.usHi;
	iStatus |= mac_addByte(pChannel, stWrite.ucPair.ucLo);
	iStatus |= mac_addByte(pChannel, stWrite.ucPair.ucHi);
	return iStatus;
}

/* volume,v amplitude [adjustment]
 -----------------------------------------------------------------------------*/
int32_t macCom_volume(MacroBlock_t *pMacro, MacroChannel_t *pChannel)
{
	char *pszVolTok,*pszAdjTok;
	int16p16_t stVol,stAdj;
	int8p8_t stWrite;

	int32_t iStatus;


	pszVolTok = strtok(NULL, szMAC_WhiteSpace);
	pszAdjTok = strtok(NULL, szMAC_WhiteSpace);
	if(NULL == pszVolTok)
		return 1;
	if(parse16p8(&stVol, pszVolTok))
		return 1;

	if(NULL != pszAdjTok)
	{
		if(parse16p8(&stAdj, pszAdjTok))
		{
			return 1;
		}
	}
	else
	{
		stAdj.iWhole = 0;
	}

	iStatus = mac_addByte(pChannel, CORETONE_PATCH_VOLUME);
	stWrite.usWhole = stVol.usPair.usHi;
	iStatus |= mac_addByte(pChannel, stWrite.ucPair.ucLo);

	stWrite.usWhole = stAdj.usPair.usLo;
	iStatus |= mac_addByte(pChannel, stWrite.ucPair.ucLo);
	stWrite.usWhole = stAdj.usPair.usHi;
	iStatus |= mac_addByte(pChannel, stWrite.ucPair.ucLo);
	return iStatus;
}


/* loop,l count
 -----------------------------------------------------------------------------*/
int32_t macCom_loop(MacroBlock_t *pMacro, MacroChannel_t *pChannel)
{
	char *pszSubTok;

	uint8_t ucCount;
	int32_t iStatus,iVal;


	if(NULL == (pszSubTok = strtok(NULL, szMAC_WhiteSpace)))
		return 1;
	if(strtoint(pszSubTok, &iVal))
		return 1;
	if((iVal < MACRO_PARSER_LOOP_MIN) || (iVal > MACRO_PARSER_LOOP_MAX))
	{
		fprintf(stderr, "%s", szERR_Range);
		return 1;
	}

	ucCount = iVal & 0xFF;
	iStatus = mac_addByte(pChannel, CORETONE_PATCH_LOOP_START);
	iStatus |= mac_addByte(pChannel, ucCount);
	return iStatus;
}

/* endloop,el
 -----------------------------------------------------------------------------*/
int32_t macCom_endloop(MacroBlock_t *pMacro, MacroChannel_t *pChannel)
{
	return mac_addByte(pChannel, CORETONE_PATCH_LOOP_END);
}

/* wait,w ticks
 -----------------------------------------------------------------------------*/
int32_t macCom_wait(MacroBlock_t *pMacro, MacroChannel_t *pChannel)
{
	char *pszSubTok;

	uint8_t aucDelay[4];
	int32_t iVal,iDiv;

	int32_t iStatus;
	uint32_t uiX,uiY;


	if(NULL == (pszSubTok = strtok(NULL, szMAC_WhiteSpace)))
		return 1;
	if(strtoint(pszSubTok, &iVal))
		return 1;
	if(iVal < MACRO_PARSER_WAIT_MIN)
	{
		fprintf(stderr, "%s", szERR_Range);
		return 1;
	}

	/**
	 * Delay commands are a special case and operate similarly to MIDI's
	 * varlength delays. Any command byte with its MSB set is interpreted
	 * as the start of a delay string which continues until a command
	 * with its MSB clear is encountered.
	 *
	 * The seven remaining bits of each byte in the delay string are used
	 * as the delay count itself, and are shifted into bits 7-0, 14-8,
	 * 21-15, or 28-22 of the accumulated delay count. If the delay is
	 * larger than four bytes' worth of ticks, it can be followed by
	 * a consecutive delay string.
	 *
	 * If the last command we processed was another delay, we need to
	 * backtrack the script and merge its delay count with our new one.
	 */
	if(pChannel->iWasWaiting)
	{
		iVal += pChannel->uiDelayAcc;
		pChannel->uiScriptLen -= pChannel->uiDelayLen;
	}

	iStatus = 0;
	while(iVal > 0)
	{
		memset(&aucDelay, 0, (sizeof(uint8_t) * 4));
		pChannel->uiDelayAcc = iVal;
		uiY = 1;
		if(iVal >= 0x200000)
		{
			/*---- Bits 28-22 ----*/
			iDiv = (iVal / 0x200000) & 0x7F;
			iVal -= (iDiv * 0x200000);

			aucDelay[3] = iDiv;
			uiY = 4;
		}
		if(iVal >= 0x4000)
		{
			/*---- Bits 21-15 ----*/
			iDiv = (iVal / 0x4000) & 0x7F;
			iVal -= (iDiv * 0x4000);

			aucDelay[2] = iDiv;
			if(uiY < 3) uiY = 3;
		}
		if(iVal >= 0x80)
		{
			/*---- Bits 14-8 ----*/
			iDiv = (iVal / 0x80) & 0x7F;
			iVal -= (iDiv * 0x80);

			aucDelay[1] = iDiv;
			if(uiY < 2) uiY = 2;
		}
		if(iVal > 0x00)
		{
			/*---- Bits 7-0 ----*/
			iDiv = iVal & 0x7F;
			iVal -= iDiv;

			aucDelay[0] = iDiv;
		}

		pChannel->iWasWaiting = -1;
		pChannel->uiDelayLen = uiY;
		for(uiX = 0; uiX < uiY; uiX++)
		{
			iStatus |= mac_addByte(pChannel,
				(uint8_t)(aucDelay[uiX] | CORETONE_PATCH_WAIT));
		}
	}

	return iStatus;
}

/* noteoff,n
 -----------------------------------------------------------------------------*/
int32_t macCom_noteoff(MacroBlock_t *pMacro, MacroChannel_t *pChannel)
{
	if(!(pMacro->iInstrument))
	{
		fprintf(stderr, "%s%s\n", szMAC_OffSFX, pMacro->pszName);
		return 1;
	}
	else if(pMacro->iHaveNoteOff)
	{
		fprintf(stderr, "%s%s\n", szMAC_MultiOff, pMacro->pszName);
		return 1;
	}

	pMacro->uiNoteOff = pChannel->uiScriptLen;
	pMacro->iHaveNoteOff = -1;
	return 0;
}

/* end,e
 -----------------------------------------------------------------------------*/
int32_t macCom_end(MacroBlock_t *pMacro, MacroChannel_t *pChannel)
{
	return mac_addByte(pChannel, CORETONE_PATCH_END);
}


/******************************************************************************
 * !!!!----                     COMMAND TABLE                        ----!!!!
 ******************************************************************************/
char *apszMacCom_nameLong[] ={
	"sample", "mode", "frequency", "volume",
	"loop", "endloop", "wait", "rest",
	"noteoff", "end",
	NULL
};

char *apszMacCom_nameShort[] = {
	"s", "m","f","v",
	"l","el","w","r",
	"n","e",
	NULL
};

MacroCom_t aMacCom_addr[]= {
	macCom_sample, macCom_mode, macCom_frequency, macCom_volume,
	macCom_loop, macCom_endloop, macCom_wait, macCom_wait,
	macCom_noteoff, macCom_end,
	NULL
};

int32_t aiMacCom_bodyOnly[] = {
	0,0,0,0,
	-1,-1,-1,-1,
	-1,-1,
	0
};


/******************************************************************************
 * !!!!----             MACRO INSTRUMENT / SFX PARSING               ----!!!!
 ******************************************************************************/
/* int32_t mac_parseBody(MacroBlock_t *pMacro, char *pszMainTok, int32_t *piBodyOpen)
 *  Handle parsing of a single instrument or sound effect once its label and
 * parsing mode (single or multi channel) have been identified.
 *  If any errors are encountered, a nonzero value will be returned.
 *----------------------------------------------------------------------------*/
int32_t mac_parseBody(MacroBlock_t *pMacro, char *pszMainTok, int32_t *piBodyOpen)
{
	int32_t iStatus;
	int32_t iBodyOpen,iAmInstrument;
	uint32_t uiX,uiY;

	/**
	 *  Before actually handling the header and body decoding we need to make
	 * a quick check to see if we're at the end of one channel's script and
	 * moving into the next one's header.
	 *
	 * Essentially:
	 *  -We're not parsing the body
	 *  -The channel has been closed (hit '}')
	 *
	 *  If this is the case, we'll set up the next channel's data stream.
	 * Note that the multimode parser in mac_parse is actually responsible
	 * for closing the sound effect based upon whether it is using single or
	 * multichannel syntax.
	 *  However, we will be checking to see if an instrument has been given
	 * more than one channel.
	 */
	iStatus = 0;
	iBodyOpen = *piBodyOpen;
	iAmInstrument = pMacro->iInstrument;

	uiX = pMacro->uiChannels;
	if(!iBodyOpen && (pMacro->pChannels[uiX].iAmClosed))
	{
		if(iAmInstrument)
		{
			fprintf(stderr, "%s%s,%u\n",
				szMAC_InsMultiC, pszMacroFile, uiMacroLine);
			return -1;
		}
		pMacro->uiChannels++;
		uiX++;

		/**
		 * Alright, the channel is valid and we've closed its entry
		 * in the macro block. Now we need to prep the next slot.
		 */
		if(tealloc(&(pMacro->pChannels), &(pMacro->uiChannelSlots),
			(pMacro->uiChannels + 1), sizeof(MacroChannel_t)))
		{
			return -1;
		}

		pMacro->pChannels[uiX].pacStream = 
			malloc(sizeof(uint8_t) * MACRO_PARSER_MACSTREAM_LEN);
		pMacro->pChannels[uiX].iAmClosed = 0;
		pMacro->pChannels[uiX].iGotSample = 0;
		pMacro->pChannels[uiX].iWasWaiting = 0;
		pMacro->pChannels[uiX].uiScriptLen = 0;
		pMacro->pChannels[uiX].uiBufferLen = MACRO_PARSER_MACSTREAM_LEN;
		if(NULL == pMacro->pChannels[uiX].pacStream)
		{
			fprintf(stderr, "%s", szERR_NoMainMem);
			return -1;
		}
	}

	/**
	 * Unless we're opening or closing the channel body with a curlybrace,
	 * we can assume the current main token is a command and we'll look it
	 * up in the command table.
	 */
	if(iBodyOpen && (0 == strcmp(pszMainTok, "}")))
	{
		if(!(pMacro->pChannels[uiX].iGotSample))
		{
			fprintf(stderr, "%s%s,%u\n", szMAC_PickSample,
				pMacro->pszName, uiX);
		}
		if(!(pMacro->iHaveNoteOff) && pMacro->iInstrument)
		{
			fprintf(stderr, "%s%s\n", szMAC_SetOff,
				pMacro->pszName);
		}

		iBodyOpen = 0;
		pMacro->pChannels[uiX].iAmClosed = -1;
	}
	else if(!iBodyOpen && (0 == strcmp(pszMainTok, "{")))
	{
		iBodyOpen = -1;
	}
	else
	{
		/**
		 * Not all commands are supported all the time, there's context
		 * involved. For example, a loop can only appear in the script body
		 * and "noteoff" only applies when parsing in instrument.
		 *
		 * The former type of cases will be handled here in dispatch.
		 */
		for(uiY = 0; (NULL != aMacCom_addr[uiY]); uiY++)
		{
			if((0 == strcmp(pszMainTok, apszMacCom_nameLong[uiY])) ||
			   (0 == strcmp(pszMainTok, apszMacCom_nameShort[uiY])))
			{
			   break;
			}
		}

		if(NULL == aMacCom_addr[uiY])
		{
			fprintf(stderr, "%s%s in : %s,%u\n", szMAC_NoCom,
				pszMainTok, pszMacroFile, uiMacroLine);
			return -1;
		}
		if(!iBodyOpen && aiMacCom_bodyOnly[uiY])
		{
			fprintf(stderr, "%s%s in : %s,%u\n", szMAC_NoFlow,
				pszMainTok, pszMacroFile, uiMacroLine);
			return -1;
		}


		if(macCom_wait != aMacCom_addr[uiY])
		{
			pMacro->pChannels[uiX].iWasWaiting = 0;
		}
		iStatus = (*aMacCom_addr[uiY])(pMacro, &(pMacro->pChannels[uiX]));
		if(iStatus > 0)
		{
			fprintf(stderr, "%s%s,%u\n", szERR_Syntax,
				pszMacroFile, uiMacroLine);
		}
	}

	*piBodyOpen = iBodyOpen;
	return iStatus;
}

/* int32_t mac_parse(FILE *pSASS, char *pszFileName, int32_t iDoSFX, int32_t iNoSamples)
 *  Parse the SASS script in the file pSASS as either a instrument script
 * (iDoSFX == 0) or sound effect script (iDoSFX != 0). As with the music parser,
 * the file's name must be stored in pszFileName in order to generate
 * helpful errors. Will skip the resolution of samples if iNoSamples is set,
 * which is useful for just loading macro names.
 *  If any issues are encountered during parsing, a nonzero value will
 * be returned.
 *----------------------------------------------------------------------------*/
int32_t mac_parse(FILE *pSASS, char *pszFileName, int32_t iDoSFX, int32_t iNoSamples)
{
	MacroBlock_t *pCurMacro;
	int32_t iNewPri = 0;

	int32_t iAborted = -1,iStatus = 0;
	int32_t iFoundMAC = 0,iMultiOpen = 0,iSingleOpen = 0;
	int32_t iBodyOpen = 0,iNewBody = 0;

	char *pszMainTok = NULL,*pszSubTok = NULL;

	/**
	 * !!!!---- PARSE SETUP ----!!!!
	 * Set up the globals required for parsing, clear out the temporary use
	 * macro block container, and allocate our macro block table.
	 *
	 * In regards to the macro block table, we're always keeping it one entry
	 * larger than the number of compiled macros (uiMacros), and this extra
	 * entry is used as the working slot for the macro currently being
	 * compiled. Channels within a macro work the same way.
	 */
	pszMacroFile = pszFileName;
	uiMacroLine = 0;
	giNoSamples = iNoSamples;

	pMacros = malloc(sizeof(MacroBlock_t) * MACRO_PARSER_MACBUF_LEN);
	if(NULL == pMacros)
	{
		fprintf(stderr, "%s", szERR_NoMainMem);
		return -1;
	}
	uiMacros = 0;
	uiMacroSlots = MACRO_PARSER_MACBUF_LEN;

	pCurMacro = pMacros + 0;
	if(mac_prepEntry(pCurMacro, !iDoSFX))
	{
		return -1;
	}


	/**
	 * !!!!---- PARSE START ----!!!!
	 * Allocate our linebuffer then begin traversing the SASS source
	 * line by line. Macro instruments and sound effects do not require
	 * multiple passes, so this function should only be called once.
	 */
	fprintf(stdout, "%s%s\n", szMAC_Parse, pszFileName);
	if(NULL == (pszMacroBuf = malloc(sizeof(char) * MACRO_PARSER_LINEBUF_LEN)))
	{
		fprintf(stderr, "%s", szERR_NoMainMem);
		return -1;
	}

	for(uiMacroLine = 1; ; uiMacroLine++)
	{
		/**
		 * Fetch our current line from the source file, stripping out any
		 * ;comments and tokenizing the remaining characters.
		 */
		if(fgets(pszMacroBuf, MACRO_PARSER_LINEBUF_LEN, pSASS) == NULL)
		{
			iAborted = 0;
			break;
		}
		if(strcut(pszMacroBuf, ";", '\0')) break;
		pszMainTok = strtok(pszMacroBuf, szMAC_WhiteSpace);

		if(iFoundMAC && (NULL != pszMainTok))
		{
			/**
			 * !!!!---- INSTRUMENT/SFX MULTIMODE PARSER ----!!!!
			 * -------------------------------------------------
			 * The original SASS specification defined both instruments and
			 * sound effects as one channel in length due to the limits of
			 * its target (Atari Lynx).
			 *
			 * We're following the new multichannel standard here (introduced
			 * in HuSound), where sound effects may be as many channels long
			 * as there are resources in the target...
			 *
			 * BUT- instruments are still limited to only a one channel width
			 * even though they may to follow the new multichannel syntax.
			 */
			if(iMultiOpen)
			{
				/**
				 * !!!!---- MULTICHANNEL PARSING ----!!!!
				 * --------------------------------------
				 * Woah, new thing: multiple channels are added to a single
				 * sound effect by enclosing all their headers and bodies in
				 * curlybraces which open right after the sound effect's name.
				 */
				if(!iBodyOpen && (0 == strcmp(pszMainTok, "}")))
				{
					/**
					 * The macro block has been closed, add it to the current
					 * table and prep the next slot.
					 */
					if(tealloc(&pMacros, &uiMacroSlots, (uiMacros + 2), sizeof(MacroBlock_t)))
						break;

					uiMacros++;
					pCurMacro = pMacros + uiMacros;
					if(mac_prepEntry(pCurMacro, !iDoSFX))
						break;

					iFoundMAC = 0;
				}
				else
				{
					/**
					 * Otherwise continue to parse the macro body...
					 */
					if(mac_parseBody(pCurMacro, pszMainTok, &iBodyOpen))
						break;
				}
			}
			else
			{
				/**
				 * !!!!---- MODE DETECTION AND SINGLE CHANNEL PARSING ----!!!!
				 * -----------------------------------------------------------
				 * Since the user could be writing either a single or
				 * multichannel definition here, we'll make the call based upon
				 * if we see '{' before any valid commands.
				 *
				 * If we see an opening curlybrace first, then this is (very)
				 * likely to be a multichannel sound effect definition. If it's
				 * one of our valid channel header commands, then we're
				 * probably decoding something in the old single channel syntax.
				 */
				if(iSingleOpen || (0 != strcmp(pszMainTok, "{")))
				{
					/**
					 * !!!!---- SINGLE CHANNEL PARSING ----!!!!
					 * ----------------------------------------
					 * Just like in the old SASSy days, only the sound
					 * effect or instrument body is curlybraced.
					 */
					if(mac_parseBody(pCurMacro, pszMainTok, &iBodyOpen))
						break;

					iSingleOpen = !(pCurMacro->pChannels[0].iAmClosed);
					if(!iSingleOpen)
					{
						if(tealloc(&pMacros, &uiMacroSlots, (uiMacros + 2), sizeof(MacroBlock_t)))
							break;

						uiMacros++;
						pCurMacro = pMacros + uiMacros;
						if(mac_prepEntry(pCurMacro, !iDoSFX))
							break;

						iFoundMAC = 0;
					}
				}
				else
				{
					if(!iSingleOpen)
					{
						iMultiOpen = -1;
					}
					else
					{
						fprintf(stderr, "%s%s,%u\n", szMAC_Unclear,
							pszFileName, uiMacroLine);
						break;
					}
				}
			}
		}
		else if(NULL != pszMainTok)
		{
			/**
			 * !!!!---- INSTRMENT / SFX DETECTION ----!!!!
			 * -------------------------------------------
			 * We're not currently parsing an instrument or sound effect,
			 * so we need to look for the start of one with the format
			 * name [priority].
			 */
			pszSubTok = strtok(NULL, szMAC_WhiteSpace);
			if(NULL != pszSubTok)
			{
				/**
				 * Priorities should just be an integer value after the sound
				 * effect name. However, they're not allowed for instruments
				 * since their priorities are taken from the currently
				 * decoding music track.
				 */
				if(1 == sscanf(pszSubTok, "%d", &iNewPri))
				{
					if((iNewPri < MACRO_PARSER_PRIORITY_MIN)
						|| (iNewPri > MACRO_PARSER_PRIORITY_MAX))
					{
						fprintf(stderr, "%s%s,%u\n", szMAC_BadPri,
							pszFileName, uiMacroLine);
						break;
					}

					if(!iDoSFX)
					{
						fprintf(stderr, "%s%s,%u\n", szMAC_InstPri,
							pszFileName, uiMacroLine);
						break;
					}

					pCurMacro->iPriority = iNewPri;
				}
				else
				{
					fprintf(stderr, "%s%s,%u\n", szMAC_BadLabel,
						pszFileName, uiMacroLine);
					break;
				}
			}
			else if((NULL == pszSubTok) && iDoSFX)
			{
				fprintf(stderr, "%s%s,%u\n", szMAC_NoPri,
					pszFileName, uiMacroLine);

				pCurMacro->iPriority = 127;
			}

			/**
			 * Check to see if the instrument or sound effect already exists.
			 * If not, the temporary use macro block should have already been
			 * set up for us. So we can assign it the new name.
			 */
			if(0 == mac_get(pszMainTok, NULL, NULL))
			{
				fprintf(stderr, "%s%s in : %s,%u\n", szMAC_TwoLabels,
					pszMainTok, pszFileName, uiMacroLine);
				break;
			}

			pCurMacro->pszName = malloc(sizeof(char) * (strlen(pszMainTok) + 1));
			if(NULL == pCurMacro->pszName)
			{
				fprintf(stderr, "%s", szERR_NoMainMem);
				break;
			}
			strcpy(pCurMacro->pszName, pszMainTok);

			iFoundMAC = -1;
			iMultiOpen = 0;
			iSingleOpen = 0;
			iBodyOpen = 0;
		}

		/**
		 * Regardless of mode, check for garbage at the end of the current
		 * parsing line before advancing.
		 */
		pszSubTok = strtok(NULL, szMAC_WhiteSpace);
		if(NULL != pszSubTok)
		{
			fprintf(stderr, "%s%s,%u\n", szERR_Garbage,
				pszFileName, uiMacroLine);
			break;
		}
	}

	/**
	 * Check to see if we bailed during the parsing of an instrument or
	 * sound effect and add a general error notification in if we broke
	 * in the middle of a macro block parse.
	 *
	 * Any work in progress macros will be freed in mac_close() if we
	 * hit a problem.
	 */
	if(iFoundMAC)
	{
		fprintf(stderr, "%s%s,%u\n", szMAC_Busy,
			pszFileName, uiMacroLine);
	}
	else if(~(iAborted | iFoundMAC) && (0 == uiMacros))
	{
		fprintf(stderr, "%s", szMAC_Empty);
	}

	return (iAborted | iFoundMAC);
}

/* void mac_close(void)
 *  Close the macro instrument / sound effect descriptor table. Call once
 * prior to the exit of the program.
 *----------------------------------------------------------------------------*/
void mac_close(void)
{
	uint32_t uiX,uiY;

	if(NULL != pszMacroBuf)
	{
		free(pszMacroBuf);
		pszMacroBuf = NULL;
	}

	if(NULL != pMacros)
	{
		for(uiX = 0; uiX <= uiMacros; uiX++)
		{
			if(NULL != pMacros[uiX].pszName)
			{
				free(pMacros[uiX].pszName);
				pMacros[uiX].pszName = NULL;
			}

			if(NULL != pMacros[uiX].pChannels)
			{
				for(uiY = 0; uiY <= pMacros[uiX].uiChannels; uiY++)
				{
					if(NULL != pMacros[uiX].pChannels[uiY].pacStream)
					{
						free(pMacros[uiX].pChannels[uiY].pacStream);
						pMacros[uiX].pChannels[uiY].pacStream = NULL;
					}
				}

				free(pMacros[uiX].pChannels);
				pMacros[uiX].pChannels = NULL;
			}
		}

		free(pMacros);
		pMacros = NULL;
	}
}




/******************************************************************************
 * !!!!----             SOUND EFFECT SPECIFIC FUNCTIONS              ----!!!!
 ******************************************************************************/
/* int32_t sfx_make(void)
 *  Attempt to build and write multiple sound effects for the currently
 * populated macro block table. Output files will be created in the same
 * directory as the original macro script, one per definition.
 *  Returns a nonzero value if anything went wrong during the build.
 *----------------------------------------------------------------------------*/
int32_t sfx_make(void)
{
	FILE *pDest = NULL;
	uint8_t *pPack = NULL;
	uint32_t *pPackDir;

	char *pszDir = NULL,*pszCurFile = NULL;
	uint32_t uiLen,uiChannels;

	int32_t iStatus = -1;
	uint32_t uiC,uiX,uiY,uiZ;

	/**
	 * !!!!---- PREBUILD ACTIVITIES ----!!!!
	 * Take a guess at the directory our macro script was located in, we'll be
	 * using this as the base for all of our output sound effects.
	 *
	 * If the macro script was located in the current working directory,
	 * pszDir will never be allocated.
	 */
	if(NULL == pszMacroFile)
	{
		fprintf(stderr, "%s", szERR_Null);
		return -1;
	}

	uiZ = 0;
	uiC = strlen(pszMacroFile);
	for(uiX = (uiC - 1); (uiX != 0xFFFFFFFF); uiX--)
	{
		for(uiY = 0; ('\0' != szMAC_DirMarkers[uiY]); uiY++)
		{
			if(pszMacroFile[uiX] == szMAC_DirMarkers[uiY])
			{
				uiZ = uiX;
				break;
			}
		}

		if(uiZ > 0)
		{
			if(NULL == (pszDir = malloc(sizeof(char) * (uiC + 1))))
			{
				fprintf(stderr, "%s", szERR_NoMainMem);
				return -1;
			}

			strcpy(pszDir, pszMacroFile);
			pszDir[(uiZ + 1)] = '\0';
			break;
		}
	}

	/**
	 * !!!!---- BUILD + WRITE BINARIES ----!!!!
	 * Generate the final binary for each of the sound effects and save them
	 * as individual files with the *.sfx extension.
	 *
	 * I'm making the usual assumption that whatever we malloc will be word
	 * aligned and excluding that, the platform we're running on can at least
	 * catch and correct misaligned writes (x86 can with a speed penalty).
	 */
	for(uiX = 0; uiX < uiMacros; uiX++)
	{
		/**
		 * The value uiChannels in the macro block descriptor is actually
		 * the INDEX of the last channel rather than the total count, so
		 * we need to add one to get the actual number of channels to build.
		 */
		uiChannels = pMacros[uiX].uiChannels + 1;
		uiZ = CORETONE_SFXPAK_HEAD_SIZE
			+ (uiChannels * CORETONE_SFXPAK_ENTRY_LEN * sizeof(uint32_t));
		uiLen = uiZ;
		for(uiY = 0; uiY < uiChannels; uiY++)
		{
			uiLen += pMacros[uiX].pChannels[uiY].uiScriptLen;
		}

		if(NULL == (pPack = malloc(sizeof(uint8_t) * uiLen)))
		{
			fprintf(stderr, "%s", szERR_NoMainMem);
			goto cleanup;
		}
		pPackDir = (uint32_t*)(pPack + CORETONE_SFXPAK_HEAD_SIZE);


		strcpy(pPack, CORETONE_SFXPAK_HEAD_MAGICWORD);
		memcpy((pPack + CORETONE_SFXPAK_HEAD_COUNT), &uiChannels, sizeof(uint32_t));
		for(uiY = 0; uiY < uiChannels; uiY++)
		{
			pPackDir[CORETONE_SFXPAK_ENTRY_SAMPLE] =
				pMacros[uiX].pChannels[uiY].uiSample;
			pPackDir[CORETONE_SFXPAK_ENTRY_SCRIPT] =
				uiZ;
			memcpy((pPack + uiZ), pMacros[uiX].pChannels[uiY].pacStream,
				(pMacros[uiX].pChannels[uiY].uiScriptLen * sizeof(uint8_t)));

			uiZ += pMacros[uiX].pChannels[uiY].uiScriptLen;
			pPackDir += CORETONE_SFXPAK_ENTRY_LEN;
		}


		/**
		 * Alright, we've prepped the binary in pPack with length uiLen. Now we
		 * need to generate the output file name (format : macro.sfx) and
		 * attempt to open it for writing.
		 */
		uiZ = strlen(szMAC_SfxExt);
		if(NULL == pszDir)
		{
			uiY = strlen(pMacros[uiX].pszName);
			if(NULL == (pszCurFile =
				malloc(sizeof(char) * (uiY + uiZ + 1))))
			{
				fprintf(stderr, "%s", szERR_NoMainMem);
				return -1;
			}

			strcpy(pszCurFile, pMacros[uiX].pszName);
			strcpy(pszCurFile + uiY, szMAC_SfxExt);
		}
		else
		{
			uiC = strlen(pszDir);
			uiY = strlen(pMacros[uiX].pszName);
			if(NULL == (pszCurFile =
				malloc(sizeof(char) * (uiC + uiY + uiZ + 1))))
			{
				fprintf(stderr, "%s", szERR_NoMainMem);
				return -1;
			}

			strcpy(pszCurFile, pszDir);
			strcpy(pszCurFile + uiC, pMacros[uiX].pszName);
			strcpy(pszCurFile + uiC + uiY, szMAC_SfxExt);
		}

		fprintf(stdout, "%s%s\n", szSFX_Make, pszCurFile);
		if(NULL == (pDest = fopen(pszCurFile, "wb")))
		{
			fprintf(stderr, "%s%s\n", szERR_NoDest, pszCurFile);
			goto cleanup;
		}

		uiZ = fwrite(pPack, sizeof(uint8_t), uiLen, pDest);
		if(uiLen != uiZ)
		{
			fprintf(stderr, "%s%s\n", szMAC_BadWrite, pszCurFile);
			goto cleanup;
		}

		free(pPack);
		pPack = NULL;

		free(pszCurFile);
		pszCurFile = NULL;

		fflush(pDest);
		fclose(pDest);
		pDest = NULL;
	}

	/**
	 * !!!!---- WRITE PRIORITY DEFINITION FILE ----!!!!
	 */
	if(swapext(pszMacroFile, &pszCurFile, szMAC_DefExt, szMAC_DirMarkers))
	{
		fprintf(stderr, "%s", szMAC_BadDef);
		goto cleanup;
	}
	if(0 == strcmp(pszMacroFile, pszCurFile))
	{
		fprintf(stderr, "%s", szMAC_BadDef);
		goto cleanup;
	}
	if(NULL == (pDest = fopen(pszCurFile, "w")))
	{
		fprintf(stderr, "%s%s\n", szERR_NoDest, pszCurFile);
			goto cleanup;
	}

	fprintf(stdout, "\nPriority definitions saved to : %s\n", pszCurFile);
	for(uiX = 0; uiX < uiMacros; uiX++)
	{
		fprintf(pDest, "#define CPRI_%s\t\t%i\n",
			pMacros[uiX].pszName,
			pMacros[uiX].iPriority);
	}
	fflush(pDest);
	fclose(pDest);
	pDest = NULL;
	
	fprintf(stdout, "Total Sound Effects : %u\n", uiMacros);
	iStatus = 0;

cleanup:
	if(NULL != pszDir) free(pszDir);
	if(NULL != pszCurFile) free(pszCurFile);
	if(NULL != pPack) free(pPack);
	if(NULL != pDest) fclose(pDest);
	return iStatus;
}




/******************************************************************************
 * !!!!----              INSTRUMENT SPECIFIC FUNCTIONS               ----!!!!
 ******************************************************************************/
/* int32_t ins_make(FILE *pDest, char *pszFileName)
 *  Attempt to build and write an instrument package for the currently populated
 * macro block table. The destination file pDest should be open for binary
 * writing at the time this function is called.
 *  Returns a nonzero value if anything went wrong during the build.
 *----------------------------------------------------------------------------*/
int32_t ins_make(FILE *pDest, char *pszFileName)
{
	uint8_t *pPack = NULL;
	uint32_t *pDir = NULL;

	uint32_t uiScriptBase,uiOff;
	uint32_t uiLen;

	uint32_t uiX,uiY;
	int32_t iStatus = -1;

	/**
	 * !!!!---- PREBUILD ACTIVITIES ----!!!!
	 * Scan through the compiled instruments to calculate the final amount
	 * of data required for the binary, then attempt to allocate it.
	 * Luckily, all instruments only have one channel, so that's the only
	 * script we need to check.
	 */
	fprintf(stdout, "%s%s\n", szINS_Make, pszFileName);

	uiLen = CORETONE_INSPAK_HEAD_SIZE
		+ (sizeof(uint32_t) * CORETONE_INSPAK_ENTRY_LEN * uiMacros);
	uiScriptBase = uiLen;
	for(uiX = 0; uiX < uiMacros; uiX++)
	{
		uiLen += pMacros[uiX].pChannels[0].uiScriptLen;
	}

	if(NULL == (pPack = malloc(sizeof(uint8_t) * uiLen)))
	{
		fprintf(stderr, "%s", szERR_NoMainMem);
		goto cleanup;
	}

	/**
	 * !!!!---- DIRECTORY SETUP & SCRIPT COPY ----!!!!
	 */
	strcpy(pPack, CORETONE_INSPAK_HEAD_MAGICWORD);
	memcpy(pPack + CORETONE_INSPAK_HEAD_COUNT, &uiMacros, sizeof(uint32_t));
	pDir = (uint32_t*)(pPack + CORETONE_INSPAK_DIR_BASE);

	uiOff = uiScriptBase;
	for(uiX = 0; uiX < uiMacros; uiX++)
	{
		pDir[CORETONE_INSPAK_ENTRY_SAMPLE] = pMacros[uiX].pChannels[0].uiSample;
		pDir[CORETONE_INSPAK_ENTRY_SCRIPT] = uiOff;
		pDir[CORETONE_INSPAK_ENTRY_NOTE_OFF] = pMacros[uiX].uiNoteOff;
		memcpy((pPack + uiOff), pMacros[uiX].pChannels[0].pacStream,
			(sizeof(uint8_t) * pMacros[uiX].pChannels[0].uiScriptLen));

		uiOff += pMacros[uiX].pChannels[0].uiScriptLen;
		pDir += CORETONE_INSPAK_ENTRY_LEN;
	}

	uiY = fwrite(pPack, sizeof(uint8_t), uiLen, pDest);
	if(uiLen != uiY)
	{
		fprintf(stderr, "%s%s\n", szMAC_BadWrite, pszFileName);
		goto cleanup;
	}

	fprintf(stdout, "\nSuccess! %u Bytes written to : %s\n", uiY, pszFileName);
	fprintf(stdout, "Total Macro Instruments : %u\n", uiMacros);
	iStatus = 0;
	fflush(pDest);

cleanup:
	if(NULL != pPack) free(pPack);

	return iStatus;
}




/******************************************************************************
 * !!!!----          MACRO INSTRUMENT / SFX DATA MANAGEMENT          ----!!!!
 ******************************************************************************/
/* int32_t mac_addByte(MacroChannel_t *pChannel, uint8_t ucByte)
 *  Add the given byte to this macro channel's data stream, returning nonzero
 * if any errors occur.
 *----------------------------------------------------------------------------*/
int32_t mac_addByte(MacroChannel_t *pChannel, uint8_t ucByte)
{
	if(tealloc(&(pChannel->pacStream), &(pChannel->uiBufferLen),
		(pChannel->uiScriptLen + 1), sizeof(uint8_t)))
	{
		return -1;
	}

	pChannel->pacStream[pChannel->uiScriptLen] = ucByte;
	pChannel->uiScriptLen++;
	return 0;
}

/* int32_t mac_prepEntry(MacroBlock_t *pMacro, int32_t iInstrument)
 *  Prep the given macro block for use, allocating space for its channels and
 * also one script for the first channel, then defaulting all parameters.
 *----------------------------------------------------------------------------*/
int32_t mac_prepEntry(MacroBlock_t *pMacro, int32_t iInstrument)
{
	MacroChannel_t *pChannel;

	pMacro->pszName = NULL;
	pMacro->iPriority = 0;
	pMacro->iInstrument = iInstrument;

	pMacro->uiChannels = 0;
	pMacro->uiChannelSlots = 0;
	pMacro->uiNoteOff = 0;
	pMacro->iHaveNoteOff = 0;
	if(NULL == (pMacro->pChannels =
		malloc(sizeof(MacroChannel_t) * MACRO_PARSER_MACCHANNEL_LEN)))
	{
		fprintf(stderr, "%", szERR_NoMainMem);
		return -1;
	}

	pMacro->uiChannelSlots = MACRO_PARSER_MACCHANNEL_LEN;
	pChannel = pMacro->pChannels + 0;
	pChannel->pacStream =  malloc(sizeof(uint8_t) * MACRO_PARSER_MACSTREAM_LEN);
	pChannel->iAmClosed = 0;
	pChannel->iGotSample = 0;
	pChannel->iWasWaiting = 0;
	pChannel->uiScriptLen = 0;
	pChannel->uiBufferLen = MACRO_PARSER_MACSTREAM_LEN;
	if(NULL == pChannel->pacStream)
	{
		fprintf(stderr, "%s", szERR_NoMainMem);
		return -1;
	}
	return 0;
}

/* int32_t mac_get(char *pszName, uint32_t *puiNum, MacroBlock_t **ppMacro)
 *  Attempt to locate and fetch a macro instrument or sound effect with the
 * name pszName, returning its index and descriptor entry in *puiNum and
 * *ppMacro, respectively.
 *
 *  Returns a nonzero value if the entry could not be located.
 *----------------------------------------------------------------------------*/
int32_t mac_get(char *pszName, uint32_t *puiNum, MacroBlock_t **ppMacro)
{
	int32_t iFailed = -1;
	uint32_t uiX;

	if(NULL != pMacros)
	{
		for(uiX = 0; uiX < uiMacros; uiX++)
		{
			if(0 == strcmp(pszName, pMacros[uiX].pszName))
			{
				if(NULL != puiNum)
					*puiNum = uiX;
				if(NULL != ppMacro)
					*ppMacro = pMacros + uiX;

				iFailed = 0;
				break;
			}
		}
	}

	return iFailed;
}
